(**************************************************************)
(*
 *  Ensemble, (Version 0.61)
 *  Copyright 1999 Cornell University
 *  All rights reserved.
 *
 *  See ensemble/doc/license.txt for further information.
 *)
(**************************************************************)
(**************************************************************)
(* PGP: interface to PGP *)
(* Authors: Zhen Xiao, Mark Hayden, 4/97 *)
(**************************************************************)
open Util
(**************************************************************)
let name = Trace.file "PGP"
let failwith = Trace.make_failwith name
let log = Trace.log name
let log1 = Trace.log (name^"1")
(**************************************************************)
(* General Support for background computation. Should be somewhere
 * else.
 *)
type comm = 
  | Error of string
  | Reply of string

let read_sock debug alarm sock f =
  log1 (fun () -> "read_sock");
  let l = ref [] in
  let read () =
    let buf = String.create 10000 in
    let len = Hsys.read sock buf 0 (String.length buf) in
    log1 (fun () -> sprintf "%s:len=%d" debug len);
    if len = 0 then (
      Alarm.rmv_sock_recv alarm sock ;
      Hsys.close sock ;
      f (String.concat "" (List.rev !l))
    ) ;
    l := (String.sub buf 0 len) :: !l
  in
  Alarm.add_sock_recv alarm name sock (Hsys.Handler0 read)

let select_process alarm notify cmd param =
  log1 (fun () -> sprintf "%s %s" cmd param);
  let (in_read,err_read) = Hsys.background_process cmd [|"PGPPASSFD=0";"HOME=/home/hayden"|] param in
  read_sock "get_error" alarm err_read (fun s -> notify (Error s)) ;
  read_sock "get_reply" alarm in_read (fun s -> notify (Reply s))

let add_pass msg =
  let pass = Arge.check name Arge.pgp_pass in
  sprintf "%s\n%s" pass msg

(**************************************************************)
type principal   = string
type sender      = principal 
type receiver    = principal 
type cipher_text = Buf.t
type clear_text  = Buf.t

let flatten_error s =
  string_map (fun c ->
    if c = '\n' then '|' else c
  ) s

(* sign the clear_text with the sender's secret key and encrypt it
 * with receiver's public key into ASCII format. The sender's pass
 * phrase, "pass", is needed to sign the message.
 * 
 * This function uses temporary file "tin" to store "clear_text", 
 * and "tin.asc" to store the encrypted file. They would be overwritten
 * if already exist.
 *)
let encrypt sender receiver clear_text =
  let clear_text = Buf.to_string clear_text in
  let pass = Arge.check name Arge.pgp_pass in
  let clear_text = hex_of_string clear_text in
  let input = pass ^ "\n" ^ clear_text in
  let command = sprintf "pgpe -fast +batchmode=1 -r %s -u %s" receiver sender in
  let (exit0,cipher,error) = Hsys.open_process command [|"PGPPASSFD=0"|] input in
  if exit0 then (
    log (fun () -> "encrypt succeeded") ;
    Some(Buf.pad_string cipher)
  ) else (
    log (fun () -> sprintf "encrypt failed") ;
    log (fun () -> sprintf "pgp info=%s" (flatten_error error)) ;
    None
  )

(* The same -- in the background
 *)
let bckgr_encrypt sender receiver clear_text alarm rep_fun =
  let clear_text = Buf.to_string clear_text in
  let pass = Arge.check name Arge.pgp_pass in
  let clear_text = (hex_of_string clear_text) in 
  let input = pass ^ "\n" ^ clear_text in
  let command = sprintf "pgpe -fast +batchmode=1 -r %s -u %s" receiver sender in
  (* Ignore stderr.
   *)
  select_process alarm (function
    | Error e -> ()
    | Reply cipher -> 
	if cipher="" then (
	  log (fun () -> "[bckground] encrypt failed") ;
	  rep_fun None
	) else (
	  log (fun () -> "[background] encrypt succeded");
	  rep_fun (Some(Buf.pad_string cipher))
	)
  ) command input
  

(* decrypt the cipher_text with my secret key. My pass phrase, "pass"
 * is needed to decrypt the message. 
 *)

(* Open: Neither "sender" nor "receiver" is actually needed. First,
 * "sender" is not needed because it doesn't matter who sent the
 * cipher_text as long as decryption is concerned. Second, "receiver"
 * is not needed because the decryption is always done using the
 * user's pass phras.  
 * 
 * This function uses temporary file "tout.asc" to store "cipher_text",
 * and "tout" to store the decrypted file. They would be overwritten if 
 * already exist.
*)

let decrypt sender receiver cipher_text =
  let cipher_text = Buf.to_string cipher_text in
  let command = sprintf "pgpe -f +batch %s -u %s" receiver sender in
  let (exit0,clear,error) = Hsys.open_process command  [|"PGPPASSFD=0"|] cipher_text in
  if exit0 then (
    log (fun () -> "decrypt succeeded") ;
    let clear = hex_to_string clear in
    Some(Buf.pad_string clear)
  ) else (
    log (fun () -> "decrypt failed") ;
    log (fun () -> sprintf "pgp info=%s" (flatten_error error)) ;
    None
  )

(* The same -- in the background
 *)
let bckgr_decrypt sender receiver cipher_text alarm rep_fun =
  let cipher_text = Buf.to_string cipher_text in
(*  let command = sprintf "pgpv -f +batch %s -u %s" receiver sender in*)
  let command = sprintf "pgpv -f +batch" in
  select_process alarm (function
    (* Ignore stderr.
     *)
    | Error e -> ()
    | Reply clear -> 
	if clear="" then (
	  log (fun () -> "[bckground] decrypt failed") ;
	  rep_fun None
	) else (
	  log (fun () -> "[background] decrypt succeded");
	  let clear = hex_to_string clear in
	  rep_fun (Some(Buf.pad_string clear))
	)
  ) command (add_pass cipher_text)


(* The following two functions are for debugging purpose. 
 * They are not part of this module.
 * 
 * Here I assume the pass phrase for "xiao" and "hayden" are just their
 * respective user id.
 *)

(*
let prstr s =
  match decrypt "xiao" "hayden" s with
  | Some s -> printf "%s\n" s
  | None -> printf "error\n"

let _ =  
 match encrypt "xiao" "hayden" "pgp is fun" with
 | Some s -> prstr s
 | None -> printf "error\n"
*)

let _ =
  let princ id nm =
    if id <> Addr.Pgp then 
      failwith sanity ;
    Addr.PgpA(nm)
  in

  let seal id src dst clear =
    let src = Addr.project src id in
    let dst = Addr.project dst id in
    match src,dst with
    | Addr.PgpA(src),Addr.PgpA(dst) ->
	encrypt src dst clear
    | _,_ -> None
  in

  let bckgr_seal id src dst clear alarm rep_fun = 
    let src = Addr.project src id in
    let dst = Addr.project dst id in
    match src,dst with
    | Addr.PgpA(src),Addr.PgpA(dst) ->
	bckgr_encrypt src dst clear alarm rep_fun
    | _,_ -> rep_fun None
  in

  let unseal id src dst cipher =
    let src = Addr.project src id in
    let dst = Addr.project dst id in
    match src,dst with
    | Addr.PgpA(src),Addr.PgpA(dst) ->
	decrypt src dst cipher
    | _,_ -> None
  in

  let bckgr_unseal id src dst cipher alarm rep_fun = 
    let src = Addr.project src id in
    let dst = Addr.project dst id in
    match src,dst with
    | Addr.PgpA(src),Addr.PgpA(dst) ->
	bckgr_decrypt src dst cipher alarm rep_fun
    | _,_ -> rep_fun None
  in

  let auth =
    Auth.create
      name
      princ
      seal
      bckgr_seal
      unseal
      bckgr_unseal
  in
  
  Auth.install Addr.Pgp auth

(**************************************************************)
