(**************************************************************)
(*
 *  Ensemble, (Version 0.61)
 *  Copyright 1999 Cornell University
 *  All rights reserved.
 *
 *  See ensemble/doc/license.txt for further information.
 *)
(**************************************************************)
(**************************************************************)
(* TRANSPORT.ML *)
(* Author: Mark Hayden, 7/95 *)
(**************************************************************)
open Util
open Trans
open View
(**************************************************************)
let name = Trace.file "TRANSPORT"
let failwith = Trace.make_failwith name
let log = Trace.log name
(**************************************************************)

let array_split3 abc =
  let a = Arrayf.map (fun (x,_,_) -> x) abc in
  let b = Arrayf.map (fun (_,x,_) -> x) abc in
  let c = Arrayf.map (fun (_,_,x) -> x) abc in
  (a,b,c)

(**************************************************************)

type dest =
  | XSend of rank
  | XCast
  | XGossip
  | XMerge of View.id option * Endpt.full

let string_of_dest = function
  | XSend rank -> sprintf "Send(%d)" rank
  | XCast -> "Cast"
  | XGossip -> "Gossip"
  | XMerge _ -> "Merge(...)"

(**************************************************************)

type 'msg t = {
  disable : unit -> unit ;
  xmit	  : dest -> 'msg
}

(**************************************************************)

let disable m 	= m.disable ()
let send m rank = m.xmit (XSend rank)
let cast m 	= m.xmit XCast
let gossip m 	= m.xmit XGossip
let merge m v e = m.xmit (XMerge(v,e))

(**************************************************************)

let find_modes ranking addr addrs = function
  | XSend dest ->
      let dest = Arrayf.get addrs dest in
      let modes = Addr.modes_of_view (Arrayf.doubleton addr dest) in
      let modes = Arrayf.filter Addr.has_pt2pt modes in
      if Arrayf.length modes = 0 then
        failwith "route:Send:no pt2pt modes in address" ;
      Arrayf.singleton (Addr.prefer ranking modes)
  | XCast ->
      let modes = Addr.modes_of_view addrs in
      let modes = Arrayf.filter Addr.has_pt2pt modes in
      if Arrayf.length modes = 0 then
        failwith "route:Cast:no pt2pt modes in address" ;
      Arrayf.singleton (Addr.prefer ranking modes)
  | XGossip ->                (* all available cast modes *)
      let modes = Addr.modes_of_view (Arrayf.singleton addr) in
      let modes = Arrayf.filter Addr.has_mcast modes in
      if Arrayf.length modes = 0 then
        failwith "route:Gossip:no mcast modes in address" ;
      modes
  | XMerge(_,dest) ->
      let modes = Addr.modes_of_view (Arrayf.doubleton addr (snd dest)) in
      let modes = Arrayf.filter Addr.has_pt2pt modes in
      if Arrayf.length modes = 0 then
        failwith "route:Merge:no pt2pt modes in address" ;
      Arrayf.singleton (Addr.prefer ranking modes)

(**************************************************************)

let find_dests endpt group addrs view = function
  | XSend(dest) ->
      let dest = Arrayf.get addrs dest in
      Domain.Pt2pt(Arrayf.singleton dest)
  | XCast ->
      (* Send to all the addresses but mine.
       *)
      if Arrayf.length addrs <> Arrayf.length view then
	failwith "mismatched view and address" ;
      let addrs = Arrayf.combine view addrs in
      let addrs = Arrayf.filter (fun (e,a) -> e <> endpt) addrs in
      let addrs = Arrayf.map snd addrs in
      Domain.Pt2pt(addrs)
  | XGossip -> Domain.Gossip(group)
  | XMerge(_,dest) -> Domain.Pt2pt(Arrayf.singleton (snd dest))

(**************************************************************)

let get_xmits alarm modes dests =
  let xmits =
    Arrayf.map (fun mode ->
      let domain = Elink.domain_of_mode alarm mode in
      Domain.xmit domain mode dests
    ) modes
  in
  Arrayf.filter_nones xmits

(**************************************************************)

let route ranking alarm endpt addr group view addrs dest =
  log (fun () -> sprintf "endpt=%s,group=%s,dest.0=%s"
    (Endpt.string_of_id endpt)
    (Group.string_of_id group)
    (string_of_dest dest)) ;

  (* Find modes to use.
   *)
  let modes = find_modes ranking addr addrs dest in

  log (fun () -> sprintf "modes=%s" (Arrayf.to_string Addr.string_of_id modes)) ;

  (* Calculate the destinations.
   *)
  let dests = find_dests endpt group addrs view dest in

  log (fun () -> sprintf "dests.1=%s" (Domain.string_of_dest dests)) ;

  (* For point-to-point:
   * (I've already been stripped)
   * 1. Strip local destinations
   * 2. Strip duplicate destinations
   * 3. Determine if any destinations are local.
   * 4. Use Mcast if we have it and there is more than 1
   *    effective destination (but not if the mode is Udp).
   *
   * For gossip: set local.
   *)
  let local,dests = match dests with
  | Domain.Pt2pt(dests) ->
      let local,dests = Addr.compress addr dests in
      let dests = 
	if Arrayf.length dests > 1
	&& modes <> Arrayf.singleton Addr.Udp
	&& Arrayf.for_all Addr.has_mcast modes
	then (
	  let loopback = true in
	  Domain.Mcast(group,loopback)
        ) else (
	  Domain.Pt2pt dests
        )
      in (local,dests)
  | Domain.Gossip _ -> true,dests
  | _ -> false,dests
  in

  log (fun () -> sprintf "local=%b, dests.2=%s" local (Domain.string_of_dest dests)) ;

  (* Collect all xmits.
   *)
  let xmits = get_xmits alarm modes dests in

  log (fun () -> sprintf "nxmits=%d" (Arrayf.length xmits)) ;

  (* Add in local xmits, if needed.  They are put at the end
   * so that messages are sent on the wire first.
   *)
  let xmits =
    if local then (
      let local = Alarm.local_xmits alarm in
      Arrayf.append xmits (Arrayf.singleton local)
    ) else (
      xmits
    )
  in

  (* Split them apart and "merge" them.
   *)
  let x,xv,xvs = array_split3 xmits in
  let x   = Route.merge3 x in
  let xv  = Route.merge1iov xv in
  let xvs = Route.merge2iovr xvs in
  (x,xv,xvs)

(**************************************************************)

(*
let enabled = 
  let table = Hashtbl.create name 100 in
  Trace.install_root (fun () ->
    [sprintf "TRANSPORT:#enabled=%d" (hashtble_size table)]
  ) ;
  table
*)
(*
    Hashtbl.iter (fun id () ->
      eprintf "  %s\n" (Conn.string_of_id_field id)
    ) enabled
*)

let f2 alarm ranking (ls,vs) stack_id router gen_recv =

  (* Create record of connections.
   *)
  let conns = Conn.create ls.rank vs.version vs.group ls.view_id stack_id vs.proto_id vs.view in

  (* Check for security problems.
   *)
  Route.security_check router ;

  (* Transmit a message.
   *)
  let xmit dest =
    let conn = match dest with
    | XCast -> 
	Conn.multi_send conns
    | XSend dest -> 
        Conn.pt2pt_send conns dest
    | XGossip -> 
	Conn.gossip conns
    | XMerge(view_id,dest) ->
	Conn.merge_send conns view_id (fst dest)
    in

    log (fun () -> sprintf "conn=%s" (Conn.string_of_id conn)) ;

    (* Determine how the message will be sent.
     *)
    let xmits = route ranking alarm ls.endpt ls.addr vs.group vs.view vs.address dest in

    (* In the end, pass everything off to router's blast function.
     *)
    Route.blast router xmits vs.key conn
  in

  (* Disable this transport instance.
   *)
  let disable () =
    let modes = Addr.ids_of_set ls.addr in
(*
    Hashtbl.remove enabled conns.Conn.id ;
*)
    Arrayf.iter (fun mode ->
      if Addr.has_mcast mode || Addr.has_pt2pt mode then (
      	let domain = Elink.domain_of_mode alarm mode in
      	Domain.disable domain mode vs.group ls.addr vs.view
      )
    ) modes ;
    Route.remove router (Alarm.handlers alarm) (Conn.key conns) (Conn.all_recv conns (Route.scaled router))
  in

  let t = {
    xmit    = xmit ;
    disable = disable 
  } in

  (* This is a hack for the bypass protocols.
   *)
  let handler,ret = gen_recv t in

  (* Enable all potential domains.
   *)
  let modes = Addr.ids_of_set ls.addr in
  Arrayf.iter (fun mode ->
    if Addr.has_mcast mode || Addr.has_pt2pt mode then (
      let domain = Elink.domain_of_mode alarm mode in
      Domain.enable domain mode vs.group ls.addr vs.view
    )
  ) modes ;
(*
    Hashtbl.add enabled conns.Conn.id () ;
*)
  Route.install router (Alarm.handlers alarm) (Conn.key conns) (Conn.all_recv conns (Route.scaled router)) vs.key handler ;
  ret

let f alarm ranking vf i j receive =
   f2 alarm ranking vf i j (fun trans -> (receive,trans))

(**************************************************************)
