(**************************************************************)
(*
 *  Ensemble, (Version 0.61)
 *  Copyright 1999 Cornell University
 *  All rights reserved.
 *
 *  See ensemble/doc/license.txt for further information.
 *)
(**************************************************************)
(**************************************************************)
(* EXCHANGE.ML : key exchange protocol *)
(* Authors: Ohad Rodeh 4/98 *)
(* Based on code by Mark Hayden, with suggestions by Ron Minsky *)
(**************************************************************)
(* Improvments
 * -----------
 * (1) Include signing of messages.
*)

open Buf
open Layer
open View
open Event
open Util
open Trans
open Shared

(**************************************************************)
let name = Trace.filel "EXCHANGE"
let failwith = Trace.make_failwith name
(**************************************************************)

(* Something that completely identifies this endpoint
 *)
type id = Trans.rank * Endpt.id * Trans.ltime * Addr.set 
  
type exch_msg = 
| Id of id 
| Ticket of id * id * Auth.ticket
    
type bckgr = 
| Idle
| Computing
| CTicket of id * Auth.ticket option
| OpenTicket of (Addr.set * Auth.clear) option 
| Accepted
      
let string_of_bckgr = function
| Idle -> "Idle"
| Computing -> "Computing" 
| CTicket _ -> "CTicket"
| OpenTicket _ -> "OpenTicket"
| Accepted -> "Accepted"
      
type state = {
  policy             : (Addr.set -> bool) option ;
  sim                : bool ;
  alarm              : Alarm.t ;
  mutable key_sug    : Security.key ;
  mutable blocked    : bool;
  mutable bckgr      : bckgr
}
	       
let dump = ignore2
	     
let check_policy s a =
  match s.policy with
  | None -> true
  | Some f -> f a
	
let init s (ls,vs) = {
  alarm   = Elink.alarm_get_hack ();
  sim     = Param.bool vs.params "exc_sim";
  key_sug = vs.key ;
  blocked = false ;
  policy  = s.exchange;
  bckgr   = Idle
}
			 
(**************************************************************)

let marshal, unmarshal = Buf.make_marsh "EXCHANGE" true

let hdlrs s ((ls,vs) as vf) {up_out=up;upnm_out=upnm;dn_out=dn;dnlm_out=dnlm;dnnm_out=dnnm}  =
  let log = Trace.log2 name ls.name in
  let logs = Trace.log2 (name^"S") ls.name in
  let logk = Trace.log2 (name^"K") ls.name in

  let shared = Cipher.lookup "RC4" in

  (* Need to add signature of the message via Digest+encryption via
   * s.key_sug. 
   *)
  let gossip_send msg =
    let msg = marshal msg in
    let msg = Buf.to_string msg in
    let msg = ExchangeGos msg in
    msg
  in
    
  (* Need to add verification of the message via Digest+encryption via
   * s.key_sug. 
   *)
  let gossip_get a = 
    let a = Buf.of_string name a in
    let src = unmarshal a len0 (Buf.length a) in 
    src
  in
    
  let gossip_recv msg = 
    let msg = gossip_get msg in
    if s.blocked || s.bckgr <> Idle then
      log (fun () -> sprintf "blocked++, %s" (string_of_bckgr s.bckgr))
    else
      match msg with 
      | Id(s_rank, s_endpt, s_lt, s_addr) -> 
	  if s_endpt <= ls.endpt then 
	    log (fun () -> sprintf "discarding id %s" 
		(Endpt.string_of_id s_endpt))
	  else
	    if not (check_policy s s_addr) then 
	      log (fun () -> "recvd id, but policy rejected it")
	    else (
	      s.bckgr <- Computing;
	      log (fun () -> sprintf "recvd and ok'd id") ;
	      Auth.bckgr_ticket s.sim ls.addr s_addr None
		(Security.buf_of_key s.key_sug) s.alarm 
		(fun t -> s.bckgr <- CTicket ((s_rank,s_endpt,s_lt,s_addr),t))
	    )
      | Ticket((my_rank,my_endpt,my_ltime,my_addr), (s_rank, s_endpt, s_lt, s_addr), ticket) -> 
	  if s_endpt = ls.endpt then 
	    log (fun () -> "Ticket from myself, discarding")
	  else
	    if not (my_rank = ls.rank 
		&& my_endpt = ls.endpt 
		&& my_ltime = vs.ltime
		&& my_addr = ls.addr) then 
	      log (fun () -> sprintf "received ticket with old my_nonce %d,%s,%d"
		  my_rank (Endpt.string_of_id my_endpt) my_ltime)
	    else (
	      log (fun () -> sprintf "got ticket from %s" 
		  (Endpt.string_of_id s_endpt));
	      s.bckgr <- Computing;
	      Auth.bckgr_check s.sim ls.addr None
		ticket s.alarm (fun clear -> s.bckgr <- OpenTicket clear)
	    )
  in
  
  let up_hdlr ev abv () = up ev abv
  and uplm_hdlr ev hdr = failwith "bad uplm event"
  and upnm_hdlr ev = match getType ev with
    (* Decrypt new_key with old_key
     * It is none if it was disseminated using Rekey. 
     *)
  | EView ->
      logs (fun () -> "up EView");
      let new_vs = getViewState ev in
      let new_vs = 
	match new_vs.key with 
	| Security.NoKey -> new_vs (* Dissemination using Rekey *)
	| Security.Common enc_key -> 
(*
	    logk (fun () -> sprintf "enc_key %d" (Buf.length enc_key));
*)
	    let initial = enc_key in 
	    let context = Cipher.init shared vs.key true in
	    let key = Cipher.encrypt 
			shared context enc_key len0 (Buf.length enc_key) in
(*
	    logs (fun () -> sprintf "decrypt %s with %s = %s" (Buf.break initial)
		(Security.str_of_key vs.key) key);
*)
	    View.set new_vs [Vs_key (Security.Common key)] 
      in
      s.blocked <- true ;
      upnm (set name ev[ViewState new_vs])
  | EBlock ->
      s.blocked <- true ;
      upnm ev
  | EGossipExt -> 
      if ls.am_coord then (
    	getExtendOpt ev (function
	| ExchangeGos(a) -> gossip_recv a ; true
	| _ -> false
	)
      ) ;
      upnm ev
	
  | EDump -> ( dump vf s ; upnm ev )
  | _ -> upnm ev
	
  and dn_hdlr ev abv = dn ev abv ()
			 
  and dnnm_hdlr ev = match getType ev with
    (* Encrypt new_key with old_key
     * It is none if it was disseminated using Rekey. 
     *)
  | EView ->
      logs (fun () -> "dn EView");
      let new_vs = getViewState ev in
      let new_vs = 
	match new_vs.key with 
        | Security.NoKey -> new_vs (* Dissemination using Rekey *)
	| Security.Common key -> 
	    match s.key_sug with 
	    | Security.NoKey -> new_vs
	    | Security.Common key_sug -> 
(*
		logk (fun () -> sprintf "key_sug %d" (String.length key_sug));
	    	logs (fun () -> sprintf "encrypt %s with %s" key_sug 
		    (Security.str_of_key vs.key));
*)
		let context = Cipher.init shared vs.key true in
		let enc_key_sug = 
		  Cipher.encrypt shared context key_sug len0 (Buf.length key_sug) 
		in
(*
	    	logs (fun () -> sprintf "After encryption vs.key=%s, enc_key_sug=%s"
		    (Security.str_of_key vs.key) enc_key_sug);
*)
	    	View.set new_vs [Vs_key (Security.Common enc_key_sug)] 
      in
      dnnm (set name ev[ViewState new_vs])
	
  | EGossipExt ->
      let ev =
	if ls.am_coord && not s.blocked then (
	  let msg = gossip_send(Id(ls.rank,ls.endpt,vs.ltime,ls.addr)) in
	  set name ev [msg]
	) else ev
      in
      dnnm ev
	
  | ETimer -> (
      if not s.blocked then
	match s.bckgr with
	| Idle 
	| Computing -> ()
	| CTicket ((s_rank,s_endpt,s_lt,s_addr), t) ->
   	    s.bckgr <- Idle;
	    (match t with 
	    | Some ticket ->
    	      	log (fun () -> sprintf "sending ticket to %s" 
      	      	    (Endpt.string_of_id s_endpt)) ;
	  	let msg = gossip_send(
	    	  Ticket((s_rank, s_endpt, s_lt, s_addr),
		  (ls.rank,ls.endpt,vs.ltime,ls.addr),
		  ticket)) in 
		dnnm (create name EGossipExt[msg]);
	    | None ->
	      	log (fun () -> sprintf "failed to authenticate id")
	    )
	| OpenTicket clear -> 
	    logs (fun () -> "OpenTicket");
	    s.bckgr <- Idle;
	    (match clear with
	    | Some(princ,key) -> 
	    	if check_policy s princ	then (
	    	  log (fun () -> sprintf "recvd and ok'd ticket") ;
		  let sec_key = Security.Common key in 
		  if sec_key <> s.key_sug then (
		    s.bckgr <- Accepted;
		    log (fun () -> "accepting key") ;
		    s.key_sug <- sec_key ;
		    logs (fun () -> "EPrompt");
		    dnnm (create name EPrompt [])
		  )
		) else
		  logs (fun () -> "recvd ticket, but policy rejected it")
	    | None ->
	      	logs (fun () -> "Ticket is empty (None)")
	    )
	| Accepted -> ()
    )
  | _ -> dnnm ev
	
  in {up_in=up_hdlr;uplm_in=uplm_hdlr;upnm_in=upnm_hdlr;dn_in=dn_hdlr;dnnm_in=dnnm_hdlr}
       
let l args vf = Layer.hdr init hdlrs None NoOpt args vf
		  
let _ = 
  Param.default "exc_sim" (Param.Bool false) ;
  Elink.layer_install name l
    
(**************************************************************)
    
    
    
