(**************************************************************)
(*
 *  Ensemble, (Version 0.61)
 *  Copyright 1999 Cornell University
 *  All rights reserved.
 *
 *  See ensemble/doc/license.txt for further information.
 *)
(**************************************************************)
(**************************************************************)
(* MNGCHAN: manage the security channels held open. *)
(* Author: Ohad Rodeh *)
(**************************************************************)
(* This layer manages the list of open security channels. It
 * integrates Authorization, limits the amount of channels
 * held open and removes all one-sided channels.  
*) 
open Util 
open Layer 
open View 
open Event 
open Trans
(**************************************************************)
let name = Trace.filel "MNGCHAN"
(**************************************************************)

type header = 
| Current of rank list
| Twosided of rank list array
	  
type state = {
  got_chan           : Once.t ;
  policy             : (Addr.set -> bool) option ;
  saved_secchan      : (Endpt.id * (Addr.set * Security.key)) list ref ;
  im_secchan         : (Endpt.id * (Addr.set * Security.key)) list ;
  saved_rekey_cleanup: Time.t ref ;
  mutable block      : bool ;
  mutable cache      : (Endpt.id , (Addr.set * Security.key)) Hashtbl.t ;
  mutable useful     : (rank * Security.key) list ;
  current            : rank list option array ;
  limit              : int ;
  cleanup_cycle      : Time.t
} 

(**************************************************************)

let string_of_channels t =  
  let l = hashtbl_to_list t in
  string_of_list (fun (eid,(addr,key)) -> Endpt.string_of_id eid) l
    
let dump = Layer.layer_dump name (fun (ls,vs) s -> [|
  sprintf "got_chan=%s, block=%b\n" (Once.to_string s.got_chan) s.block ;
  sprintf "cache=%s\n" (string_of_channels s.cache)
|])
	
(**************************************************************)

let init s (ls,vs) = { 
  got_chan      = Once.create "got_chan";
  policy        = s.exchange ;
  saved_secchan = s.secchan;
  im_secchan    = !(s.secchan);
  saved_rekey_cleanup = s.rekey_cleanup ;
  block         = false;
  cache         = Hashtbl.create 31;        
  useful        = [];
  current       = Array.create ls.nmembers None;
  limit         = Param.int vs.params "con_number" ;
  cleanup_cycle = Param.time vs.params "cleanup_cycle"
}
		       
let check_policy s a =
  match s.policy with
    | None -> true
    | Some f -> f a
	  
(**************************************************************)
	  
let hdlrs s ((ls,vs) as vf) {up_out=up;upnm_out=upnm;dn_out=dn;dnlm_out=dnlm;dnnm_out=dnnm} =
  let failwith = layer_fail dump vf s name in
  let log = Trace.log2 name ls.name in
  let logbm = Trace.log3 "BUFFER-MNG" name ls.name in
  let logt = Trace.log2 (name^"T") ls.name in
  
(* Take hashtbl [l] and leave only the first [i] elements.
 *)
  let sub_hashtbl t i = 
    let count = ref 0 in
    Hashtbl.iter (fun x _ -> 
      incr count;
      if !count > i then Hashtbl.remove t x
    ) t
  in
  
(* Handle a leader message describing all two-sided channels. 
 * Remove onesided channels. 
 *)
  let handle_twosided tws = 
    let tws = tws.(ls.rank) in
    let u = List.map (fun (r,key) -> 
      if List.mem r tws then Some (r,key) else (
	let eid = Arrayf.get vs.view r in
	Hashtbl.remove s.cache eid;
	None
      )
    ) s.useful in
    s.useful <- filter_nones u;
    let ev = (create name EChannelList[ChannelList s.useful]) in
    upnm ev	
  in
  
(* Check if all the members sent their current sets. When
 * all channels are received check that all connections
 * are two sided. Send a "matrix" describing all twosided
 * channels.  
 *)
  let recv_all () = 	
    let is_some = function Some _ -> true | None -> false 
    in
    if array_for_all is_some s.current then (
      let a = Array.map (some_of name) s.current in
      let a = Array.mapi (fun me l -> let l = 
	List.map (fun i -> if List.mem me a.(i) then Some i else None) l in
      filter_nones l 
      ) a in
      log (fun () -> sprintf "Twosided channels=%s" (string_of_array string_of_int_list a));
      if ls.am_coord 
	&& not s.block then (
	  dnlm (create name ECast[]) (Twosided a) ;
	  handle_twosided a 
	)
    )
  in			
  
(* Send the list of connections to the leader.
 *)
  let send_current () = 
    let ranks,_ = List.split s.useful in
    if ls.am_coord then (
      s.current.(ls.rank) <- Some ranks;
      recv_all ()
    ) else
      if not s.block then dnlm (sendPeer name vs.coord) (Current ranks)
  in
  
  let up_hdlr ev abv () = up ev abv
			    
  and uplm_hdlr ev hdr = match getType ev,hdr with
    |	ESend, Current l when ls.am_coord -> 
	  let src = getPeer ev in
	  s.current.(src) <- Some l;
	  recv_all ()
	    
    (* Leader message declaring the usefull channels for this view. 
     * Send down to Secchan.
     *)
    |	ECast, Twosided tws -> handle_twosided tws
	    
    |	_ -> failwith "Sanity, unknown event type in uplm_hdlr"
	    
  and upnm_hdlr ev = match getType ev with
  (* Perform authorization and enforce a limit on the number of channels.
   * If more than the connection lifetime has expired, erase all the channels.
   *)
  | EView -> 
       if not (Once.isset s.got_chan) then 
	 failwith "Eview but haven't received a ChannelList from Secchan";
       let vs = getViewState ev in
       Hashtbl.iter (fun eid (addr,key) -> 
	 if not (check_policy s addr) then 
	   Hashtbl.remove s.cache eid 
      ) s.cache;
      sub_hashtbl s.cache s.limit;
      
      let ctime = 
	let alarm = Elink.alarm_get_hack () in
	Alarm.gettime alarm 
      in
      
      if (Time.ge ctime (Time.add s.cleanup_cycle !(s.saved_rekey_cleanup))) then	(
	log (fun () -> "Cleanup time expired, eraseing all the channels");
	s.saved_secchan := [];
	s.saved_rekey_cleanup := ctime
      ) else (
	log (fun () -> sprintf "up EView %s" (string_of_channels s.cache));
	s.saved_secchan := hashtbl_to_list s.cache
      );
      let ev = set name ev [ViewState vs] in 
      upnm ev

  (* Send to the leader the list of open Channels. Don't send the keys,
   * only the ranks of connected members. 
   *)
   | EInit -> 
       let secchan = !(s.saved_secchan) in
       let c_l = List.map (fun (eid,(addr,key)) -> 
	 try 
	   Some (Arrayf.index eid vs.view,key) 
	 with Not_found ->  None
       ) secchan in
       s.useful <- filter_nones c_l;
      send_current ();
      upnm ev
	
  | EBlock -> 
      s.block <- true;
      upnm ev
	
    (* Dump information about status of buffers.
     *)
  | EAccount ->
      logbm (fun () -> sprintf "num_channels=%d" (hashtbl_size s.cache));
      upnm ev
	
  | EDump -> ( dump vf s ; upnm ev )
	
  | _ -> upnm ev
	
  and dn_hdlr ev abv = dn ev abv ()
			 
  and dnnm_hdlr ev = match getType ev with
    (* Got a channelList from Secchan -- merge with the old channel list. 
     *)
  | EChannelList -> 
      Once.set s.got_chan;
      let c_l = getChannelList ev in
      List.iter  (function (rank,key) -> 
	Hashtbl.add s.cache	(Arrayf.get vs.view rank) 
	  (Arrayf.get vs.address rank, key)
      ) c_l;
      List.iter (fun (eid,(addr,key)) -> 
	try 
	  let _ = Hashtbl.find s.cache eid in ()
	with Not_found -> 
	  Hashtbl.add s.cache eid (addr,key)
      ) s.im_secchan ;
      free name ev
	    
    |	_ -> dnnm ev
	    
  in {up_in=up_hdlr;uplm_in=uplm_hdlr;upnm_in=upnm_hdlr;dn_in=dn_hdlr;dnnm_in=dnnm_hdlr}
       
let l args vf = Layer.hdr init hdlrs None NoOpt args vf
		  
let _ = 
  Param.default "con_number" (Param.Int 20) ;
  Param.default "cleanup_cycle" (Param.Time (Time.of_int (8 * 60 * 60))) ; (* 8 hours *)
  Elink.layer_install name l
    
(**************************************************************)
	
	
	
			

			
