(**************************************************************)
(*
 *  Ensemble, (Version 0.61)
 *  Copyright 1999 Cornell University
 *  All rights reserved.
 *
 *  See ensemble/doc/license.txt for further information.
 *)
(**************************************************************)
(**************************************************************)
(* REKEY.ML : switching to a new key using tree structure *)
(* Authors: Zhen Xiao, Mark Hayden, 6/97 *)
(* Optimization: Ohad Rodeh 12/97  *)
(**************************************************************)
open Util
open Layer
open View
open Event
open Trans
open Shared
(**************************************************************)
let name = Trace.filel "REKEY"
(**************************************************************)

type header =
  | NoHdr
  | Ack
  | Commit
  | Rekey

(**************************************************************)

type state = {
  mutable commit     : bool;
  mutable father     : rank option;
  mutable children   : rank list ;
  mutable key_sug    : Security.key ;
  mutable rekey_req  : bool ;
  rekeying           : Once.t ;
  blocking           : Once.t ;
  mutable ack        : int
} 
	       
(**************************************************************)
	       
let dump = Layer.layer_dump name (fun (ls,vs) s -> [|
  sprintf "ack=%d\n" s.ack ;
  sprintf "blocking=%s, rekeying=%s\n"
    (Once.to_string s.blocking) (Once.to_string s.rekeying)
|])

(**************************************************************)

let init s (ls,vs) = 
  let degree = Param.int vs.params "rekey_degree" in
  let first_child n = degree * n + 1 in
  let father n = 
    if n<=0 then None
    else Some ((n-1)/degree)
  in
  let children =
    let rec loop rank degree =
      if (degree = 0) || (rank >= ls.nmembers) then []
      else rank :: (loop (succ rank) (degree - 1))
    in loop (first_child ls.rank) degree
  in 
  { 
  commit    = false;
  father    = father ls.rank;
  children  = children;
  key_sug   = Security.NoKey ;
  rekey_req = false;
  rekeying  = Once.create name ;
  blocking  = Once.create "blocking" ;
  ack       = 0
}

(**************************************************************)

let hdlrs s ((ls,vs) as vf) {up_out=up;upnm_out=upnm;dn_out=dn;dnlm_out=dnlm;dnnm_out=dnnm} =
  let failwith = layer_fail dump vf s name in
  let log = Trace.log2 name ls.name in

  let all_acked () =
    if s.rekey_req
      && Once.isset s.rekeying
      && not (Once.isset s.blocking) then (
    	if s.ack = List.length s.children 
	   && not s.commit then (
      	     log (fun () -> sprintf "%d all_ack" ls.rank);
	     match s.father with 
	     | None -> 
      		 s.commit <- true;
      		 dnlm (castEv name) Commit;
      		 dnnm (create name EPrompt[])
	     | Some rank -> 
		 dnlm (sendPeer name rank) Ack
	   )
      )
  in
  
  (* Encrypt and then send the key to all its children (except the father).
   *)
  let dis_key () =
    log (fun () -> sprintf "Distributing new key to my children: %s" 
	(string_of_int_list s.children)) ;    
    let key_sug = Security.buf_of_key s.key_sug in
    List.iter (fun rank ->
      let ev = (create name ESecureMsg[SecureMsg(rank,key_sug)]) in
      dnnm ev 
    ) s.children
  in
  
  let enter_rekey () = 
    if s.rekey_req
      && not (Once.isset s.rekeying)
      && not (Once.isset s.blocking) then (
	Once.set s.rekeying;
	dis_key ()
      )
  in
  
  let init_rekey () =
    let key = Prng.create () in
    s.key_sug <- key;
    log (fun () -> sprintf "init_rekey, generating a new key");
    enter_rekey ()
  in
  
  let up_hdlr ev abv () = up ev abv
			    
  and uplm_hdlr ev hdr = match getType ev,hdr with
    
(* Some members replied that they got the key. Check whether
 * all members have the key. If so, prompt for a view change.
 *)
  | (ECast|ESend), Ack ->
      let origin = getPeer ev in 
      s.ack <- succ s.ack;
      all_acked ()
	
  | ECast, Commit -> 
      s.commit <- true;
      if ls.rank<>0 then 
	dnnm (create name EPrompt[]);
      
    (* Got a request to start rekeying the group.
     *)
  | (ECast|ESend), Rekey ->
      if ls.rank=0 then (
	s.rekey_req <- true;
      	init_rekey ()
      )
  | _ -> failwith unknown_local
	
  and upnm_hdlr ev = match getType ev with

    (* Switch to a new conversation key.  Here we assume the
     * Auth module authenticated a member when it did
     * encryption for that member.  
     *)
  | ERekey ->
      log (fun () -> sprintf "ERekey event") ;
      if ls.rank = 0 then (
	s.rekey_req <- true;
	init_rekey ()
      ) else 
      	dnlm (sendPeer name 0) Rekey;
      upnm ev
	
    (* Got a ticket from the parent.  
     *)
  | ESecureMsg when ls.rank <> 0 ->
      let _,key = getSecureMsg ev in
      s.rekey_req <- true;
      s.key_sug <- Security.Common key;
      enter_rekey () ;
      all_acked () ;
      free name ev
	
  | ESecureMsg when ls.rank = 0 ->
      failwith "Sanity, received a Sealed key when not leader";
      free name ev
	
    (* EBlock: Mark that we are now blocking.
     *)
  | EBlock ->
      Once.set s.blocking ;
      upnm ev
	
    (* EView: If all members have acknowledged the new key and no group 
     * merge occurs, then switch to using it.
     *)
  | EView ->
      let view_old = (Arrayf.to_list (vs.view)) in
      let vs_new = getViewState ev in
      let view_new = Arrayf.to_list vs_new.view in
      let no_merger = Lset.subtract view_new view_old = [] in
      if Once.isset s.rekeying 
	&& no_merger 
	&& s.commit then (
	  if ls.rank = 0 then log (fun () -> sprintf "EView:switching keys:%s" (Security.string_of_key s.key_sug)) ;
	  let vs_new = getViewState ev in
	  let vs_new = View.set vs_new [Vs_key s.key_sug; Vs_new_key true] in
	  let ev = set name ev[ViewState vs_new] in
      	  upnm ev
     	) else (
	  if (Once.isset s.rekeying) then
	    log (fun () -> sprintf "Rekey failed, merge=%s, commit=%s, coord=%d" 
		(string_of_bool (not no_merger)) (string_of_bool s.commit) vs.coord);
	  let vs_new = View.set vs_new [Vs_new_key false] in
	  let ev = set name ev [ViewState vs_new] in
	  upnm ev
	)
	  
  | EDump -> ( dump vf s ; upnm ev )
  | _ -> upnm ev
	
  and dn_hdlr ev abv = dn ev abv ()
			 
  and dnnm_hdlr = dnnm
		    
in {up_in=up_hdlr;uplm_in=uplm_hdlr;upnm_in=upnm_hdlr;dn_in=dn_hdlr;dnnm_in=dnnm_hdlr}
     
let l args vf = Layer.hdr init hdlrs None NoOpt args vf
		  
let _ = 
  Param.default "rekey_degree" (Param.Int 2) ;
  Elink.layer_install name l

(**************************************************************)
    
