(**************************************************************)
(*
 *  Ensemble, (Version 0.61)
 *  Copyright 1999 Cornell University
 *  All rights reserved.
 *
 *  See ensemble/doc/license.txt for further information.
 *)
(**************************************************************)
(**************************************************************)
(* SECCHAN: Creating secure channels between ensemble processes.*)
(* Author: Ohad Rodeh *)
(**************************************************************)
(* Fine points:
 * 1. Two members may setup channels to each other simultaniously. 
 *    To prevent this, we use a two phase protocol.
 *)
open Util
open Layer
open View
open Event
open Trans
open Shared
(**************************************************************)
let name = Trace.filel "SECCHAN"
(**************************************************************)

type header = 
  | SealedKey of Auth.ticket
  | Ack
  | Data of Security.key * Buf.t
	  
type bckgr = 
  | Idle
  | Computing
  | Ticket of rank * Auth.ticket option
  | OpenTicket of rank * (Addr.set * Auth.clear) option 
      
type request = 
  | Encrypt of rank * Buf.t
  | Decrypt of rank * Auth.ticket
	  
type con_stat = 
  | Closed
  | SemiOpen
  | Open

type channel = {
  mutable key    : Security.key;
  msgs           : Buf.t Queue.t;
  mutable status : con_stat ;
}	
		 
let string_of_con_stat = function
  | Closed  -> "Closed"
  | SemiOpen -> "SemiOpen"
  | Open    -> "Open"
	
(**************************************************************)
	
type state = {
  ready           : Once.t ;
  mutable num_ready  : int ;
  policy             : (Addr.set -> bool) option ;
  sim                : bool; (* Simulate authentication? *)
  alarm              : Alarm.t ; 
  channels           : channel option array; (* array of links *)
  mutable block      : bool ;
  mutable bckgr      : bckgr ;
  request            : request Queue.t ;
  p_sec_msgs         : (rank * Buf.t) Queue.t ;
  p_comm             : (rank * header) Queue.t 
} 
	       
(**************************************************************)
	       
let dump (ls,vs) s =
  eprintf "SECCHN:dump:%s\n" ls.name ;
  eprintf "links=%s\n" (string_of_array (string_of_option (fun c ->
    sprintf "#msgs=%d, status=%s" (Queue.length c.msgs) (string_of_con_stat c.status)))
    s.channels)
    
(**************************************************************)
    
let init s (ls,vs) = { 
  ready           = Once.create "ready" ;
  num_ready       = 0 ;
  policy          = s.exchange;
  sim             = Param.bool vs.params "sec_sim" ;
  alarm           = Elink.alarm_get_hack () ;
  channels        = Array.create ls.nmembers None;
  block           = false ;
  bckgr           = Idle ; 
  request         = Queue.create () ;
  p_sec_msgs      = Queue.create () ;
  p_comm          = Queue.create () 
}
		       
(**************************************************************)
		       
let hdlrs s ((ls,vs) as vf) {up_out=up;upnm_out=upnm;dn_out=dn;dnlm_out=dnlm;dnnm_out=dnnm} =
  let failwith = layer_fail dump vf s name in
  let log = Trace.log2 name ls.name in
  let logt = Trace.log2 (name^"T") ls.name in
  let logbm = Trace.log3 "BUFFER-MNG" ls.name name in
  
  let handle = function
    | Encrypt (dst,new_key) -> 
	  let dst_addr = Arrayf.get vs.address dst in
	  Auth.bckgr_ticket s.sim ls.addr dst_addr None	new_key s.alarm 
  	    (fun t -> s.bckgr <- Ticket (dst,t))
    | Decrypt (src,cipher_text) -> 
        Auth.bckgr_check s.sim ls.addr None
	  cipher_text s.alarm (fun clear -> s.bckgr <- OpenTicket (src,clear))
  in
  
  let push_request r = 
    if (Queue.length s.request = 0) 
      && s.bckgr = Idle then (
	s.bckgr <- Computing ;
	handle r
      ) else
	Queue.add r s.request
	  
  and pop_request () = 
    if Queue.length s.request > 0 then (
      s.bckgr <- Computing ;
      let r = Queue.take s.request in
      handle r
    )
  in
  
  let send_sealed_key dst sealed_key = 
    let c = some_of name s.channels.(dst) in
    match c.status with 
    | Closed | SemiOpen -> 
	  if c.key <> Security.NoKey 
	    && not s.block then 
	      dnlm (sendPeer name dst) (SealedKey sealed_key)
	  else () (* A Channel from the other direction is being opened *)
    | Open -> ()  (* A Channel from the other direction has been opened *)
	  
  and recv_sealed_key src key = 
    log (fun () -> sprintf "recv_sealed_key %d" src);
    let channel = {
      key    = Security.Common key ;
      msgs   = Queue.create ();
      status = Open
    }	in 
    s.channels.(src) <- Some channel;
    if not s.block then 
      dnlm (sendPeer name src) Ack
  in
  
  let handle_data peer key msg = 
    match s.channels.(peer) with
    |	None -> failwith (sprintf "Sanity, received a SecureMsg from %d before the channel is set up." peer)
    |	Some c -> 
	  if c.key <> key then 
	    failwith (sprintf "Sanity, his key (%d)<> my key" peer)
	  else
	    let e = (create name ESecureMsg[SecureMsg(peer,msg)]) in
	    upnm e
  in
  
	(* Handle a secure message [msg] to be sent to [dst]. If a
	 * channel already exists, then send/enqueue. If it does not exist, create
	 * the channel structure, enqueue the message, and create a channel.
	 *)
  let handle_secure_msg dst msg = 
    match s.channels.(dst) with 
    | None -> 
	log (fun () -> sprintf "open_chan to %d" dst);
	let channel = {
	  key    = Security.NoKey;
	  msgs   = Queue.create ();
	  status = Closed
	}	in
	Queue.add msg channel.msgs;
	let new_key = Prng.create () in
	channel.key <- new_key;
	s.channels.(dst) <- Some channel;
	push_request (Encrypt (dst, (Security.buf_of_key new_key)))
    | Some c -> 
	if not s.block then 
	  if c.status = Open then 
	    dnlm (sendPeer name dst) (Data (c.key,msg))
	  else
	    Queue.add msg c.msgs 
  in
  
	(* If his rank is lower, ignore. Otherwise try to decrypt the (SealedKey t) and
	 * establish a secure channel.
	 *)
  let handle_sealed_key peer t = 
    match s.channels.(peer) with 
    | None -> 
	push_request (Decrypt (peer,t))
    | Some c -> 
	match c.status with 
	| Closed | SemiOpen -> 
	    if peer < ls.rank then (
	      log (fun () -> sprintf "received a SealedKey from %d while opening channel" peer);
	      c.key <- Security.NoKey; (* Ignore the SealedKey operation *)
	      push_request (Decrypt (peer,t))
	    ) 
	| Open -> 
	    failwith (sprintf "Sanity, received a SealedKey from %d, although there is an open Channel already" peer)
  in	
  
	(* Received an Ack from member [peer], sends all pending messages on the queue.
	 *)
  let handle_ack peer = 
    let c = some_of name s.channels.(peer) in
    c.status <- Open;
    Queue.iter (fun msg -> dnlm (sendPeer name peer) (Data (c.key,msg))) c.msgs;
    Queue.clear c.msgs;
  in
  
	(* Received the EChannelList from Mngchan. Handle all pending messages.
	 *)
  let now_ready () = 
    Once.set s.ready ;
    Queue.iter (fun (peer,header) -> match header with 
    | SealedKey t -> handle_sealed_key peer t
    | Ack -> handle_ack peer
    | Data (key,msg) -> handle_data peer key msg
    ) s.p_comm;
    Queue.clear s.p_comm;
    Queue.iter (fun (rank,msg) -> handle_secure_msg rank msg) s.p_sec_msgs;
    Queue.clear s.p_sec_msgs
  in
  
  let up_hdlr ev abv () = up ev abv
			    
  and uplm_hdlr ev hdr = match getType ev,hdr with
  | ESend, SealedKey t -> 
      let peer = getPeer ev in
      log (fun () -> sprintf "SealedKey %d" peer);
      if (Once.isset s.ready) then
	handle_sealed_key peer t 
      else Queue.add (peer,SealedKey t) s.p_comm
	
  | ESend, Ack -> 
      let dst = getPeer ev in
      log (fun () -> sprintf "Ack %d" dst);
      if (Once.isset s.ready) then
	handle_ack dst
      else Queue.add (dst,Ack) s.p_comm
	
  | ESend, Data (key,msg) -> 
      let src = getPeer ev in
      if (Once.isset s.ready) then
	handle_data src key msg 
      else Queue.add (src,Data(key,msg)) s.p_comm
	
  | _ -> failwith "Sanity, unknown event type in uplm_hdlr"
	
  and upnm_hdlr ev = match getType ev with
	(* Received Block. Send to Mngchan the list of open channels
	 *)
  | EBlock ->
      logt (fun () -> "EBlock");
      s.block <- true;
      let c_l = Array.mapi (fun i c -> 
	match c with 
	| None -> None
	| Some {key=key;msgs=_;status=status} -> 
		  if status = Open then Some (i,key) else None
      ) s.channels in
      let c_l = filter_nones (Array.to_list c_l) in
      let cev = (create name EChannelList[ChannelList c_l]) in
      dnnm cev;
      upnm ev
	
	(* Condense the local channel list and send it up to Mngchan.
	 *)
  | ETimer -> 
      (match s.bckgr with 
      | Idle -> ()
      | Computing -> ()
      | Ticket (r,t) -> 
	  s.bckgr <- Idle;
	  (match t with 
	  | None -> failwith (sprintf "Could not create a ticket for %d" r)
	  | Some t -> send_sealed_key r t; pop_request ()
	  );
      | OpenTicket (r,clear) -> 
	  s.bckgr <- Idle;
	  match clear with 
	  | None -> failwith (sprintf "Sanity, could not authenticate ticket from %d" r)
	  | Some (_,key) -> recv_sealed_key r key; pop_request ()
      );
      upnm ev
	
	(* Got a channel list from Mngchan.
	 *)
  | EChannelList -> 
      let c_l = getChannelList ev in
      List.iter (fun (rank,key) -> 
	let c = {key=key; msgs = Queue.create (); status = Open } in
	s.channels.(rank) <- Some c
      ) c_l;
      now_ready (); 
      free name ev;
      log (fun () -> sprintf "got_channels %s" 
	  (string_of_list (fun (r,_) -> string_of_int r) c_l))
	
    (* Dump information about status of buffers.
     *)
  | EAccount ->
      let count = ref 0 in
      let a = Array.map (function 
	| None -> None 
	| Some c -> incr count; Some (Queue.length c.msgs)
      ) s.channels in
      let pending_msgs = Array.fold_left (fun x y -> match y with
	| None -> x 
	| Some y -> x+y
      ) 0 a in
      logbm (fun () -> sprintf "num_channels=%d pending_msgs=%d, p_sec_msgs=%d, p_comm=%d" 
	  !count pending_msgs (Queue.length s.p_sec_msgs) (Queue.length s.p_comm));
      upnm ev
	
  | EDump -> ( dump vf s ; upnm ev )
	
  | _ -> upnm ev
	
  and dn_hdlr ev abv = dn ev abv ()
			 
  and dnnm_hdlr ev = match getType ev with
   (* Got a secureMsg to send
    *)
  | ESecureMsg -> 
      let dst,msg = getSecureMsg ev in
      if Once.isset s.ready then 
	handle_secure_msg dst msg
      else
	Queue.add (dst,msg) s.p_sec_msgs
	  
  | _ -> dnnm ev
	
  in {up_in=up_hdlr;uplm_in=uplm_hdlr;upnm_in=upnm_hdlr;dn_in=dn_hdlr;dnnm_in=dnnm_hdlr}
       
let l args vf = Layer.hdr init hdlrs None NoOpt args vf
		  
let _ = 
  Param.default "sec_sim" (Param.Bool false) ;
  Elink.layer_install name l

(**************************************************************)
